import multiprocessing as mp
import numpy as np
from time import time
import sys
import fmq

q1 = mp.Queue(10)
q2 = fmq.Queue(10)

# a = np.zeros((100, 256, 256, 3))
a = np.zeros((1024, 720, 3), dtype=np.uint8)
a_size = sys.getsizeof(a)
print('%d bytes, %dKB, %dMB' % (a_size, a_size/1024, a_size/1024/1024))


def get_and_print(q, name):
    # mp queue get
    for i in range(5):
        st = time()
        print('Get item')
        b = q.get()
        print(f'{name} get() a time', time() - st)


# def get_and_print_fmq(q):
#     # fmq queue get
#     for i in range(5):
#         st = time()
#         b = q.get()
#         print('fmq get() a time', time() - st)


# for i in range(5):
# mp.Process(target=get_and_print, args=(q1, 'mq'), daemon=True).start()

# for i in range(5):
mp.Process(target=get_and_print, args=(q2, 'fmq'), daemon=True).start()


for i in range(5):
    print('Put item')
    # q1.put(np.array(a))
    q2.put(np.array(a))
print(q2)
while True:
    pass


import time
import threading
from .table import print_table
from .logger import get_logger

class Statistic():
    def __init__(self, funicorn_app=None):
        self.funicorn_app = funicorn_app
        self.start_time = time.time()
        self.stats_info = {
            'status': 'live',
            'uptime': 0,
            'total_req': 0,
            'total_res': 0,
            'avg_latency': 0,
            'avg_req': 0,
            'avg_res': 0,
            'crashes': 0,
        }
        self.lock = threading.Lock()
        self.logger = get_logger(name='Stat', mode='info')
        self.logger.info('Init statistics')
        
    def __getitem__(self, attr):
        return self.stats_info[attr]

    def __setitem__(self, attr, value):
        with self.lock:
            self.stats_info[attr] = value

    def increment(self, name, value=1):
        with self.lock:
            self.stats_info[name] += value

    def decrement(self, name, value=1):
        with self.lock:
            self.stats_info[name] -= value

    def average(self, name):
        with self.lock:
            return self.stats_info[name]

    @property
    def info(self):
        self.update()
        return self.stats_info

    def update(self):
        uptime = int(time.time() - self.start_time)
        with self.lock:
            self.stats_info['uptime'] = uptime
            self.stats_info['avg_req'] = 0 if uptime == 0 else round(self.stats_info['total_req']/uptime, 2)
            self.stats_info['avg_res'] = 0 if uptime == 0 else round(self.stats_info['total_res']/uptime, 2)

    @property
    def cli_info(self):
        self.update()
        table = [['status', 'uptime', 'total requests', 'total responses', 'avg latency', 'avg requests', 'avg responses', 'crashes'],
                 [self.stats_info['status'], self.stats_info['uptime'], self.stats_info['total_req'], self.stats_info['total_res'],
                  self.stats_info['avg_latency'], self.stats_info['avg_req'], self.stats_info['avg_res'], self.stats_info['crashes']]]
        return print_table(table, color=(0, 255, 0))



def find_largest_element(rows, cols, length_array, matrix):
    # Loop through each row
    for i in range(rows):
        # Loop through each column
        for j in range(cols):
            length_array.append(len(str(matrix[i][j])))
    # Sort the length matrix so that we can find the element with the longest length
    length_array.sort()
    # Store that length
    largest_element_length = length_array[-1]

    return largest_element_length


def create_matrix(rows, cols, matrix_to_work_on, matrix):
    # Loop through each row
    for i in range(rows):
        # Append a row to matrix_to_work_on for each row in the matrix passed in
        matrix_to_work_on.append([])
        # Loop through each column
        for j in range(cols):
            # Add a each column of the current row (in string form) to matrix_to_work_on
            matrix_to_work_on[i].append(str(matrix[i][j]))


def make_rows(rows, cols, largest_element_length, rowLength, matrix_to_work_on, final_table, color):
    for i in range(rows):
        current_row = ""
        for j in range(cols):
            # If we are using colors then do the same thing but as without (below)
            if ((color != None) and (i == 0)):
                current_el = " " + "\033[38;2;" + str(color[0]) + ";" + str(
                    color[1]) + ";" + str(color[2]) + "m" + matrix_to_work_on[i][j] + "\033[0m"
            # If we are not using colors (or j != 0 or i != 0) just add a space and the element that should be in that position to a variable which will store the current element to work on
            else:
                current_el = " " + matrix_to_work_on[i][j]
            # If the raw element is less than the largest length of a raw element (raw element is just the unformatted element passed in)
            if (largest_element_length != len(matrix_to_work_on[i][j])):
                # If we are using colors then add the amount of spaces that is equal to the difference of the largest element length and the current element (minus the length that is added for the color)
                # * The plus two here comes from the one space we would normally need and the fact that we need to account for a space that tbe current element already has
                if (color != None):
                    if (i == 0):
                        current_el = current_el + " " * (largest_element_length - (len(current_el) - len("\033[38;2;" + str(
                            color[0]) + ";" + str(color[1]) + ";" + str(color[2]) + "m" + "\033[0m")) + 2) + "|"
                    # If it is not the first column or first row than it doesn't need to subtract the color length
                    else:
                        current_el = current_el + " " * \
                            (largest_element_length - len(current_el) + 2) + "|"
                # If we are not using color just do the same thing as above when we were using colors for when the row or column is not the first each time
                else:
                    current_el = current_el + " " * \
                        (largest_element_length - len(current_el) + 2) + "|"
            # If the raw element length us equal to the largest length of a raw element then we don't need to add extra spaces
            else:
                current_el = current_el + " " + "|"
            # Now add the current element to the row that we are working on
            current_row += current_el
        final_table.append("|" + current_row)
    if (color != None):
        rowLength = len(current_row) - len("\033[38;2;" + str(color[0]) + ";" + str(
            color[1]) + ";" + str(color[2]) + "m" + "\033[0m")
    else:
        rowLength = len(current_row)
    return rowLength


def print_rows_in_table(final_table, print_headers=True):
    if print_headers:
        print(final_table[0])

    for row in final_table[1::]:
        print(row, end='\r', flush=True)


def print_table(matrix, color=None):
    rows = len(matrix)
    cols = len(matrix[0])
    length_array = []
    largest_element_length = None
    rowLength = None
    matrix_to_work_on = []
    final_table = []
    largest_element_length = find_largest_element(
        rows, cols, length_array, matrix)
    create_matrix(rows, cols, matrix_to_work_on, matrix)
    rowLength = make_rows(rows, cols, largest_element_length,
                          rowLength, matrix_to_work_on, final_table, color)
    return final_table



import numpy as np
import pickle
import os
import itertools
import psutil
import gc
import sys

print('NumPy version:', np.__version__, 'Python version:', sys.version)
for i in itertools.count():
    pickle.dumps(np.zeros(1024*1024))
    gc.collect()
    if i % 100 == 0:
        info = psutil.Process(os.getpid()).memory_info()
        print(f'vms={info.vms/1024/1024}MB rss={info.rss/1024/1024}MB')



import logging
import uuid
from logging import Formatter
import os

#--------------------- Logger ---------------------#

MAPPING = {
    'DEBUG': 37,  # white
    'INFO': 36,  # cyan
    'WARNING': 33,  # yellow
    'ERROR': 31,  # red
    'CRITICAL': 41,  # white on red bg
}

PREFIX = '\033['
SUFFIX = '\033[0m'


class ColoredFormatter(Formatter):
    def __init__(self, patern):
        Formatter.__init__(self, patern)

    def format(self, record):
        levelname = record.levelname
        seq = MAPPING.get(levelname, 37)  # default white
        colored_levelname = ('{0}{1}m{2}{3}').format(
            PREFIX, seq, levelname, SUFFIX)
        record.levelname = colored_levelname
        return Formatter.format(self, record)


def get_logger(name=f'logger-{str(uuid.uuid4())}', mode='debug'):
    logger = logging.getLogger(name)
    # https://stackoverflow.com/questions/17745914/python-logging-module-is-printing-lines-multiple-times
    if not logger.hasHandlers():
        handler = logging.StreamHandler()
        formatter = ColoredFormatter(
            '%(asctime)s | %(levelname)-8s | %(name)s: %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG if mode == 'debug' else logging.INFO)
    return logger


def get_fstream_logger(file_path, name=f'{str(uuid.uuid4())}', mode='debug'):
    name = 'fstream_logger-{}'.format(name)
    logger = logging.getLogger(name)
    # https://stackoverflow.com/questions/17745914/python-logging-module-is-printing-lines-multiple-times
    if not logger.hasHandlers():
        handler = logging.FileHandler(file_path)
        formatter = Formatter(
            '%(asctime)s | %(levelname)-8s | %(name)s: %(message)s')
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG if mode == 'debug' else logging.INFO)
    return logger


class Logger():
    def __init__(self, log_dir, log_prefix):
        error_file_path = os.path.join(log_dir, log_prefix + '.error')
        info_file_path = os.path.join(log_dir, log_prefix + '.info')
        self.error_logger = get_fstream_logger(
            error_file_path, name=log_prefix + '.error')
        self.info_logger = get_fstream_logger(
            info_file_path, name=log_prefix + '.info')

    def info(self, msg):
        self.info_logger.info(msg)

    def error(self, msg):
        self.error_logger.info(msg)



class LengthEqualtyError(Exception):
    pass


class ImageEncodeError(Exception):
    pass


class MaxFileSizeExeeded(Exception):
    pass


class NotSupportedInputFile(Exception):
    pass

class InitializationError(Exception):
    pass

class CommandError(Exception):
    pass


import multiprocessing as mp
import numpy as np
from time import time
import sys
import fmq

q1 = mp.Queue(10)
q2 = fmq.Queue(10)

# a = np.zeros((100, 256, 256, 3))
a = np.zeros((1024, 720, 3))
a_size = sys.getsizeof(a)
print('%d bytes, %dKB, %dMB' % (a_size, a_size/1024, a_size/1024/1024))

for i in range(5):
    q1.put(np.array(a))
    q2.put(np.array(a))

# mp queue get
for i in range(5):
    st = time()
    b = q1.get()
    print('mp get() a time', time() - st)

# fmq queue get
for i in range(5):
    st = time()
    b = q2.get()
    print('fmq get() a time', time() - st)

q2.close()


import importlib
from funicorn import Funicorn

# hello = importlib.import_module('hello')
# hello.hello()

# importlib.import_module()

model_cls = getattr(importlib.import_module('test'), 'TestModel')
model = model_cls()
model.call()


from termcolor import colored
from inspect import getfullargspec
import uuid
import time
import psutil
import cv2
import numpy as np
import sys
import importlib
import os
from PIL import Image
from io import BytesIO, StringIO

#--------------------- Image Encode/Decode ---------------------#


def img_bytes_to_img_arr(img_bytes):
    '''Convert image bytes to image array'''
    img_flatten = np.frombuffer(img_bytes, dtype=np.uint8)
    img_arr_decoded = cv2.imdecode(img_flatten, cv2.IMREAD_ANYCOLOR)
    return img_arr_decoded


def img_arr_to_img_bytes(img_arr, quality=100):
    '''Convert image array to image bytes'''
    ret, img_flatten = cv2.imencode('.jpg', img_arr, params=[
                                    cv2.IMWRITE_JPEG_QUALITY, quality])
    img_bytes = img_flatten.tobytes()
    return img_bytes


def check_tree_status(parent_pid, including_parent=True):
    status = {}
    parent = psutil.Process(parent_pid)
    for idx, child in enumerate(parent.children(recursive=True)):
        status[f'child-{idx}'] = {'pid': child.pid, 'status': child.status()}
    if including_parent:
        status['parent'] = {'pid': parent.pid, 'status': parent.status()}
    return status


def check_all_ps_status(list_pid):
    status = {}
    for pid in list_pid:
        status[pid] = check_ps_status(pid)
    return status


def check_ps_status(pid):
    ps = psutil.Process(pid)
    return ps.status()


def colored_worker_name(worker_name):
    return colored(worker_name, 'green', attrs=['bold'])


def colored_funicorn_name(funicorn_app_name='FUNICORN'):
    return colored(funicorn_app_name, 'cyan', attrs=['bold'])


def colored_network_name(network_name):
    return colored(network_name, 'magenta', attrs=['bold'])


#--------------------- CLI ---------------------#
def get_args_from_class(cls):
    init_cls_args = getfullargspec(cls)
    init_cls_args.args.remove('self')
    return init_cls_args.args


def split_class_from_path(path):
    # Append current working directory to search module
    sys.path.append(os.getcwd())
    subpaths = path.split('.')
    if len(subpaths) > 2:
        pkg, cls_name = subpaths[-2:]
    else:
        pkg, cls_name = subpaths
    cls_name = getattr(importlib.import_module(pkg), cls_name)
    return pkg, cls_name


#------------------- OTHERS ------------------#
def get_size(obj, seen=None):
    """Recursively finds size of objects"""
    size = sys.getsizeof(obj)
    if seen is None:
        seen = set()
    obj_id = id(obj)
    if obj_id in seen:
        return 0
    # Important mark as seen *before* entering recursion to gracefully handle
    # self-referential objects
    seen.add(obj_id)
    if isinstance(obj, dict):
        size += sum([get_size(v, seen) for v in obj.values()])
        size += sum([get_size(k, seen) for k in obj.keys()])
    elif hasattr(obj, '__dict__'):
        size += get_size(obj.__dict__, seen)
    elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
        size += sum([get_size(i, seen) for i in obj])
    return size



import requests
from requests.exceptions import ConnectionError
from ..exceptions import CommandError
import click
from ..funicorn import Funicorn
from ..http_api import HttpAPI
from ..rpc import ThriftAPI
from ..stat import Statistic
from ..table import print_rows_in_table
from ..utils import get_args_from_class, split_class_from_path
import importlib
import os
import sys
import time


CONTEXT_SETTINGS = {'show_default': True}

common_options = [
    click.option('-h', '--host', default='0.0.0.0', show_default=True,
                 help='funicorn service host.'),
    click.option('-p', '--port', required=True, show_default=True,
                 type=int, help='funicorn service port.')
]

funicorn_app_options = [
    click.option('--model-cls', type=str, required=True,
                 help='Model class'),
    click.option('--funicorn-cls', type=str, default=None,
                 help='Customized Funicorn class [optional]'),
    click.option('--http-cls', type=str, default=None,
                 help='Customized HTTP class [optional]'),
    click.option('--rpc-cls', type=str, default=None,
                 help='Customized RPC class [optional]'),
    click.option('--num-workers', type=int, default=1,
                 help='A number of workers'),
    click.option('--batch-size', type=int, default=1,
                 help='Inference batch size'),
    click.option('--batch-timeout', type=float,
                 default=10, help='Batch timeout (ms)'),
    click.option('--max-queue-size', type=int,
                 default=1000, help='Max queue size'),
    click.option('--http-host', type=str, default='0.0.0.0', help='HTTP host'),
    click.option('--http-port', type=int, default=5000,
                 required=True, help='HTTP port'),
    click.option('--http-threads', type=int, default=10, help='HTTP threads'),
    click.option('--rpc-host', type=str, default='0.0.0.0',
                 help='RPC (Thrift) host'),
    click.option('--rpc-port', type=int, default=None,
                 help='RPC (Thrift) port'),
    click.option('--rpc-threads', type=int, default=10,
                 help='A number of RPC threads'),
    click.option('--gpu-devices', type=str, default=None, help='GPU devices'),
    click.option('--debug', type=bool, default=False, help='debug'),
    click.argument('model-init-kwargs', nargs=-1),
]


def cli_requests(url, method='get', params=None, timeout=1):
    try:
        if method == 'get':
            resp = requests.get(url, params=params, timeout=timeout)
        elif method == 'post':
            resp = requests.post(url, timeout=timeout)
        return resp.json()
    except ConnectionError:
        print('Cannot connect to service! Service may not be started or stopped.')
        exit()


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


@click.command()
@add_options(common_options)
@click.option('--refresh', type=int, default=1, show_default=True,
              help='Refresh time')
def status(host, port, refresh=1):
    ''' View dashboard CLI
    '''
    if refresh < 1:
        refresh = 1
    is_print_header = True
    url = f'http://{host}:{port}/api/cli_status'
    while True:
        try:
            stt = cli_requests(url)
            print_rows_in_table(stt, print_headers=is_print_header)
            if is_print_header:
                is_print_header = False
            time.sleep(refresh)
        except KeyboardInterrupt:
            exit()


@click.command()
@add_options(common_options)
def worker_terminate(host, port):
    ''' Terminate all workers CLI
    '''
    url = f'http://{host}:{port}/api/terminate'
    print('> Waiting for terminate to complete...')
    stt = cli_requests(url)
    print('> ' + stt)


@click.command()
@add_options(common_options)
def worker_idle(host, port):
    ''' Idle all workers CLI
    '''
    url = f'http://{host}:{port}/api/idle'
    stt = cli_requests(url)
    print('> ' + stt)


@click.command()
@add_options(common_options)
def worker_resume(host, port):
    ''' Resume all workers CLI
    '''
    url = f'http://{host}:{port}/api/resume'
    stt = cli_requests(url)
    print('> ' + stt)


@click.command()
@add_options(common_options)
def worker_restart(host, port):
    ''' Restart all workers CLI
    '''
    url = f'http://{host}:{port}/api/restart'
    print('> Waiting for restart to complete...')
    stt = cli_requests(url)
    print('> ' + stt)


@click.command()
@add_options(common_options)
@click.option('--num-workers', type=int, default=0, show_default=True,
              help='Additional num workers')
@click.option('--gpu-devices', type=str, default=None, show_default=True, help='GPU devices')
def add_workers(host, port, num_workers, gpu_devices):
    url = f'http://{host}:{port}/api/add_workers'
    print('> Add more workers...')
    params = {'num_workers': num_workers, 'gpu_devices': gpu_devices}
    stt = cli_requests(url, params=params)
    print('> ' + stt)


@click.command(context_settings=CONTEXT_SETTINGS)
@add_options(funicorn_app_options)
def start(model_cls, funicorn_cls=None, http_cls=None, rpc_cls=None,
          num_workers=1, batch_size=1, batch_timeout=10,
          max_queue_size=1000,
          http_host='0.0.0.0', http_port=5000, http_threads=30,
          rpc_host='0.0.0.0', rpc_port=None, rpc_threads=30,
          gpu_devices=None, model_init_kwargs=None, debug=False):
    """ Welcome to Funicorn CLI.\n
        Funicorn CLI is about to help developers start Deep Learning service in the fastest way!\n

        Example:\n
            funicorn --model-cls main.Model --batch-size 4 --num-workers 4 --gpu-devices 1,2,3            
        
        Additionally, developers can try:\n
            - funicorn-add: Add more model workers.\n
            - funicorn-idle: Idle all model workers.\n
            - funicorn-resume: Resume all model workers.\n
            - funicorn-terminate: Terminate all model workers.\n
            - funicorn-status: View the service's dashboard .\n
    """
    if funicorn_cls is None:
        funicorn_cls = Funicorn
    else:
        pkg, funicorn_cls = split_class_from_path(funicorn_cls)
    if model_cls is not None:
        pkg, model_cls = split_class_from_path(model_cls)

        if model_init_kwargs:
            model_init_kwargs = dict(kwarg.split(':')
                                     for kwarg in model_init_kwargs)

    if gpu_devices:
        gpu_devices = [
            gpu_id for gpu_id in gpu_devices.split(',') if gpu_id != '']

    funicorn_app = funicorn_cls(model_cls=model_cls,
                                num_workers=num_workers,
                                batch_size=batch_size,
                                batch_timeout=batch_timeout,
                                max_queue_size=max_queue_size,
                                gpu_devices=gpu_devices,
                                model_init_kwargs=model_init_kwargs,
                                debug=debug)

    stat = Statistic(funicorn_app=funicorn_app)

    if (rpc_host and http_host) and (rpc_port == http_port):
        raise ConnectionError('rpc_port and http_port must be different.')

    if http_port:
        if http_cls is not None:
            pkg, http_cls = split_class_from_path(http_cls)
            assert issubclass(http_cls, HttpAPI)
        else:
            http_cls = HttpAPI
        http = http_cls(funicorn_app=funicorn_app, stat=stat,
                        host=http_host, port=http_port,
                        threads=rpc_threads,
                        debug=debug)
    if rpc_port:
        if rpc_cls is not None:
            pkg, rpc_cls = split_class_from_path(rpc_cls)
            assert issubclass(rpc_cls, ThriftAPI)
        else:
            rpc_cls = ThriftAPI
        rpc = rpc_cls(funicorn_app=funicorn_app, stat=stat,
                      host=rpc_host, port=rpc_port,
                      threads=rpc_threads,
                      debug=debug)
    funicorn_app.serve()



from funicorn import Funicorn, HttpAPI
from funicorn import ThriftAPIV2
from funicorn import Statistic
from funicorn.logger import get_logger
import os
import tensorflow as tf


class TestModel():
    def __init__(self, gpu_id, model_path):
        self.logger = get_logger()

    def predict(self, batch):
        batch_len = len(batch)
        result = []
        for _ in range(batch_len):
            result.append({'val_shape': 3})
        return result

    def postprocess(self, result):
        return result

if __name__ == '__main__':
    app = Funicorn(TestModel,
                   batch_size=1, batch_timeout=0.01,
                   num_workers=2, gpu_devices=[1, 1, 1, 4, 5],
                   model_init_kwargs={'model_path': 'path'})
    stat = Statistic(funicorn_app=app)
    http_api = HttpAPI(funicorn_app=app, stat=stat)
    thrift_api = ThriftAPIV2(funicorn_app=app, host='0.0.0.0', stat=stat, port=5005, threads=2)
    
    app.serve()



from funicorn import Funicorn
from funicorn.thriftcorn.rpc_api import ThriftAPI
from funicorn.utils import colored_funicorn_name, colored_network_name
import random
from nlpservice.nlpservice import NLPService


class NLPThriftApi(ThriftAPI):
    def __init__(self, *args, **kwargs):
        ThriftAPI.__init__(self, name='NLPThriftApi', *args, **kwargs)
        self.logger.name = colored_network_name('NLPThriftApi')

    def init_processor(self, handler):
        processor = NLPService.Processor(handler)
        return processor


class NLPGateWay(Funicorn):
    '''Customize Funicorn Service to work as waygate'''

    def __init__(self, *args, **kwargs):
        Funicorn.__init__(self, *args, **kwargs)
        self.entries = [
            'Kiki-Handler-V1', 
            'Kiki-Handler-V2'
        ]
        self.logger.name = colored_funicorn_name('NLPGateWay')

    def get_entries(self, idx):
        return self.entries[idx]

    def nlp_encode(self, text):
        '''Distribute text to other entries'''

        # Logic to distribute
        entry = self.get_entries(random.randint(0, len(self.entries) - 1))
        self.logger.info(f'Send {text} to {entry} and wait the result')
        # Maybe send to other Funicorn Service Entry

        # Then wait for the result
        # In this case we return a dummy result
        self.logger.info(f'Receive {text} from {entry}')
        result = text
        return result



from thrift.server import TNonblockingServer
from thrift.protocol import TBinaryProtocol
from thrift.transport import TSocket, TTransport

from .FunicornService import Processor
from .thrift_server import TModelPool
from ..logger import get_logger
from ..utils import colored_network_name
import threading
import time
import json
import uuid


class Handler():
    def __init__(self, funicorn_app, stat, logger):
        self.funicorn_app = funicorn_app
        self.stat = stat
        self.logger = logger
        self.logger.info('Init Handler!')

    def preprocess(self, data):
        return data

    def predict_img_bytes(self, img_bytes):
        start_time = time.time()
        assert isinstance(img_bytes, bytes)
        data = self.preprocess(img_bytes)
        json_result = self.funicorn_app.predict(data)
        self.stat.increment('total_req')
        if isinstance(json_result, str) or isinstance(json_result, dict):
            ValueError('The result from rpc must be json string')
        self.logger.info(f'process-time: {time.time() - start_time}')
        self.stat.increment('total_res')
        return json.dumps(json_result)

    def ping(self):
        self.logger.info('Ping!')


class ThriftAPI(threading.Thread):
    def __init__(self, funicorn_app, host, port, name='RPC', stat=None, threads=40,
                 timeout=1000, debug=False):
        threading.Thread.__init__(self, daemon=True)
        self.name = name
        self.funicorn_app = funicorn_app
        self.host = host
        self.port = port
        self.stat = stat
        self.threads = threads
        self.debug = debug
        self.logger = get_logger(colored_network_name(
            'RPC'), mode='debug' if debug else 'info')
        self.funicorn_app.register_connection(self)

    def init_connection(self, processor):
        '''
        Initialize all connections with TBinaryProtocol and TFramedTransport

        Parameters
        ----------
        processor : TProcessor
            Processor class has been generated by thrift

        Returns
        -------
        server: TServer
            TNonblocking Server
        '''
        socket = TSocket.TServerSocket(host=self.host, port=self.port)
        prot_fac = TBinaryProtocol.TBinaryProtocolFactory()
        server = TNonblockingServer.TNonblockingServer(processor=processor,
                                                       lsocket=socket,
                                                       inputProtocolFactory=prot_fac,
                                                       threads=self.threads)
        return server

    def init_handler(self):
        return Handler(self.funicorn_app, self.stat, self.logger)

    def init_processor(self, handler):
        processor = Processor(handler)
        return processor

    def run(self):
        handler = self.init_handler()
        processor = self.init_processor(handler)
        server = self.init_connection(processor)
        self.logger.info(
            f'Server is running at http://{self.host}:{self.port}')
        server.serve()


class ThriftAPIV2(ThriftAPI):
    def __init__(self, handler_cls=None, processor_cls=None, *args, **kwargs):
        ThriftAPI.__init__(self, *args, **kwargs)
        self._handler_cls = Handler if handler_cls is None else handler_cls
        self._processor_cls = Processor if processor_cls is None else processor_cls

    @property
    def handler_cls(self):
        return self._handler_cls

    @handler_cls.setter
    def handler_cls(self, handler_cls):
        self._handler_cls = handler_cls

    @property
    def processor_cls(self, processor_cls):
        return self._processor_cls

    @processor_cls.setter
    def processor_cls(self, processor_cls):
        self.processor_cls = processor_cls

    def run(self):
        socket = TSocket.TServerSocket(host=self.host, port=self.port)
        prot_fac = TBinaryProtocol.TBinaryProtocolFactory()
        list_model_config = [
            {"funicorn_app": self.funicorn_app,
             "stat": self.stat,
             "logger": self.logger}]*self.threads
        server = TModelPool.TModelPoolServer(
            handler_cls=self._handler_cls,
            processor_cls=self._processor_cls,
            tsocket=socket, protocol_factory=prot_fac,
            list_model_config=list_model_config
        )
        self.logger.info(
            f'Server is running at http://{self.host}:{self.port}')
        server.serve()



import sys
from termcolor import colored, cprint

text = colored('Hello, World!', 'red', attrs=['reverse', 'blink'])
print(text)
cprint('Hello, World!', 'green', 'on_red')


def print_red_on_cyan(x): return cprint(x, 'red', 'on_cyan')


print_red_on_cyan('Hello, World!')
print_red_on_cyan('Hello, Universe!')

for i in range(10):
    cprint(i, 'magenta', end=' ')

cprint("Attention!", 'red', attrs=['bold'], file=sys.stderr)



from funicorn.logger import get_logger
import os
import tensorflow as tf

class NSFW_V2():
    def __init__(self, gpu_id, model_path):
        self.logger = get_logger()
        self.graph = tf.Graph()
        self.sess = tf.Session(graph=self.graph)
        with self.graph.as_default():
            self.x = tf.ones(shape=(3, 3), dtype=tf.float32)
            self.y = tf.ones(shape=(3, 3), dtype=tf.float32)

    def predict(self, batch):
        batch_len = len(batch)
        result = []
        for _ in range(batch_len):
            result.append({'val_shape': 3})
        return result

    def postprocess(self, result):
        return result




from funicorn import Funicorn
from funicorn.logger import get_logger
import os
import tensorflow as tf


class TestModel():
    def __init__(self, gpu_id, model_path):
        self.logger = get_logger()
        self.graph = tf.Graph()
        self.sess = tf.Session(graph=self.graph)
        with self.graph.as_default():
            self.x = tf.ones(shape=(3, 3), dtype=tf.float32)
            self.y = tf.ones(shape=(3, 3), dtype=tf.float32)

    def predict(self, batch):
        batch_len = len(batch)
        result = []
        for _ in range(batch_len):
            result.append({'val_shape': 3})
        return result

    def postprocess(self, result):
        return result


# if __name__ == '__main__':
#     funicorn = Funicorn(TestModel,
#                         http_host='0.0.0.0', http_port=8000,
#                         rpc_host='0.0.0.0', rpc_port=8001,
#                         batch_size=1, batch_timeout=0.01,
#                         num_workers=1, gpu_devices=[1, 1, 1, 4, 5],
#                         model_init_kwargs={'model_path': 'path'})
#     funicorn.serve()



from flask import Flask, request, abort, jsonify
from waitress import serve
import threading
from PIL import Image
import numpy as np
import time
from collections import namedtuple
from http import HTTPStatus
import traceback

from .exceptions import NotSupportedInputFile, MaxFileSizeExeeded, InitializationError
from .utils import colored_network_name, check_all_ps_status
from .logger import get_logger
from .stat import Statistic


class HttpAPI(threading.Thread):
    def __init__(self, funicorn_app, host='0.0.0.0', port=5001, stat=None, threads=40, name='HTTP', timeout=1000, debug=False, register_conn=True):
        threading.Thread.__init__(self, daemon=True)
        self.name = name
        self.host = host
        self.port = port
        self.threads = threads
        self.timeout = timeout
        self.funicorn_app = funicorn_app
        self.stat = stat or Statistic(funicorn_app=funicorn_app)
        self.logger = get_logger(colored_network_name('HTTP'),
                                 mode='debug' if debug else 'info')
        if register_conn:
            self.funicorn_app.register_connection(self)

    def init_exception(self, app):
        @app.errorhandler(HTTPStatus.REQUEST_ENTITY_TOO_LARGE)
        def max_file_size_exeeded(error):
            resp = jsonify({
                "error_code": HTTPStatus.REQUEST_ENTITY_TOO_LARGE,
                "error_message": 'Request Size Exeeded',
                "results": []
            })
            resp.status_code = HTTPStatus.REQUEST_ENTITY_TOO_LARGE
            return resp

        @app.errorhandler(HTTPStatus.NOT_FOUND)
        def not_found(error):
            resp = jsonify({
                "error_code": HTTPStatus.NOT_FOUND,
                "error_message": "Api Not Found",
                "results": []
            })
            resp.status_code = HTTPStatus.NOT_FOUND
            return resp

        @app.errorhandler(HTTPStatus.BAD_REQUEST)
        def wrong_request_params(error):
            resp = jsonify({
                "error_code": HTTPStatus.BAD_REQUEST,
                "error_message": "Wrong request parameter",
                "results": []
            })
            resp.status_code = HTTPStatus.BAD_REQUEST
            return resp

        @app.errorhandler(HTTPStatus.INTERNAL_SERVER_ERROR)
        def internal_server_error(error):
            resp = jsonify({
                "error_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                "error_message": "Internal server error",
                "results": []
            })
            resp.status_code = HTTPStatus.INTERNAL_SERVER_ERROR
            return resp
        return app

    def create_app(self):
        app = Flask(__name__)
        app = self.init_exception(app)

        def check_request_size(request, max_size=5 * 1024 * 1024):
            if request.content_length > max_size:
                raise MaxFileSizeExeeded(
                    "Input file size too large, limit is {:0.2f}MB".format(max_size/(1024**2)))

        def convert_bytes_to_img_arr(img_bytes):
            try:
                img = Image.open(img_bytes).convert("RGB")
                img = np.array(img, np.uint8)
                return img
            except:
                raise NotSupportedInputFile(
                    "Wrong input file type, only accept image")

        @app.route("/api/predict_img_bytes", methods=['POST'])
        def predict_img_bytes():
            final_res = []
            try:
                self.stat.increment('total_req')
                check_request_size(request)
                if 'img_bytes' in request.files:
                    img_bytes = request.files['img_bytes']
                    img = convert_bytes_to_img_arr(img_bytes)
                    results = self.funicorn_app.predict_img_bytes(img)
                    if results is not None:
                        final_res = results
                    else:
                        abort(HTTPStatus.BAD_REQUEST)
                    final_res = final_res if len(final_res) != 0 else []
                    resp = jsonify({
                        "error_code": 0,
                        "error_message": "Successful.",
                        "results": final_res
                    })
                    resp.status_code = HTTPStatus.OK
                    self.stat.increment('total_res')
                    return resp
                else:
                    self.stat.increment('crashes')
                    abort(HTTPStatus.BAD_REQUEST)

            except NotSupportedInputFile as e:
                self.stat.increment('crashes')
                abort(HTTPStatus.BAD_REQUEST)

            except MaxFileSizeExeeded as e:
                self.stat.increment('crashes')
                abort(HTTPStatus.REQUEST_ENTITY_TOO_LARGE)

            except Exception as e:
                self.stat.increment('crashes')
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)

        @app.route('/api/predict_json', methods=['POST'])
        def predict_json():
            self.stat.increment('num_req')
            try:
                url = request.args['url']
                result = self.funicorn_app.predict(url)
                self.stat.increment('num_res')
                self.logger.info(f'result is: {result}')
            except Exception as e:
                return jsonify({'result': e})
            else:
                return jsonify({'result': result})

        @app.route('/api/status', methods=['GET'])
        def status():
            try:
                resp = jsonify(self.stat.info)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/cli_status', methods=['GET'])
        def cli_status():
            try:
                resp = jsonify(self.stat.cli_info)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/resume', methods=['GET'])
        def resume_all_workers():
            try:
                ps_stt = self.funicorn_app.resume_all_workers()
                resp = jsonify(ps_stt)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/idle', methods=['GET'])
        def idle_all_workers():
            try:
                ps_stt = self.funicorn_app.idle_all_workers()
                resp = jsonify(ps_stt)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/terminate', methods=['GET'])
        def terminate_all_workers():
            try:
                ps_stt = self.funicorn_app.terminate_all_workers()
                resp = jsonify(ps_stt)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/restart', methods=['GET'])
        def restart_all_workers():
            try:
                ps_stt = self.funicorn_app.restart_all_workers()
                resp = jsonify(ps_stt)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/add_workers', methods=['GET'])
        def add_workers():
            try:
                num_workers = request.args.get('num_workers')
                gpu_devices = request.args.get('gpu_devices')
                gpu_devices = gpu_devices.split(',') \
                    if gpu_devices is not None else None
                ps_stt = self.funicorn_app.add_more_workers(
                    num_workers, gpu_devices)
                resp = jsonify(ps_stt)
                resp.status_code = HTTPStatus.OK
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
            else:
                return resp

        @app.route('/api/predict_url', methods=['POST', 'GET'])
        def predict_url():
            final_res = []
            try:
                self.stat.increment('total_req')
                if 'url' in request.args:
                    url = request.args['url']
                    results = self.funicorn_app.predict(url)
                    if results is not None:
                        final_res = results
                    else:
                        abort(HTTPStatus.INTERNAL_SERVER_ERROR)
                    resp = jsonify({
                        "error_code": HTTPStatus.OK,
                        "error_message": "Successful.",
                        "results": final_res
                    })
                    resp.status_code = HTTPStatus.OK
                    self.stat.increment('total_res')
                    return resp
                else:
                    # self.stat.increment('crashes')
                    abort(HTTPStatus.BAD_REQUEST)

            except NotSupportedInputFile as e:
                # self.stat.increment('crashes')
                abort(HTTPStatus.BAD_REQUEST)

            except MaxFileSizeExeeded as e:
                self.stat.increment('crashes')
                abort(HTTPStatus.REQUEST_ENTITY_TOO_LARGE)

            except Exception as e:
                self.stat.increment('crashes')
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)

        @app.route('/', methods=['GET'])
        def index():
            try:
                return 'Welcome to Funicorn'
            except Exception as e:
                self.logger.error(traceback.format_exc())
                abort(HTTPStatus.INTERNAL_SERVER_ERROR)
        # End of API
        return app

    # https://docs.pylonsproject.org/projects/waitress/en/stable/arguments.html#arguments
    def run(self):
        self.flask_app = self.create_app()
        if self.funicorn_app is None:
            raise InitializationError(
                'Cannot start HTTP service. Funicorn app is required when start http service!')
        self.logger.info(
            f'Service is running on http://{self.host}:{self.port}')
        serve(app=self.flask_app,
              host=self.host, port=self.port,
              threads=self.threads, _quiet=True, backlog=1024)



import time
import sys
import cv2
import numpy as np

from funicorn.rpc.ttypes import *
from funicorn.rpc import FunicornService
from funicorn.utils import (img_bytes_to_img_arr,
                            img_arr_to_img_bytes,
                            colored_worker_name)
from funicorn.logger import get_logger
from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.transport import TTransport, TSocket


class ClientRPC():
    '''Simple Client for Funicorn'''

    def __init__(self, port, host='0.0.0.0', timeout_ms=1000, debug=False, asynchronous=False):
        self.host = host
        self.port = port
        self.transport = None
        self.client = None
        self.asynchronous = asynchronous
        self.results = {}
        self.logger = get_logger(colored_worker_name('CLIENT'),
                                 mode='debug' if debug else 'info')

    def get_connection(self):
        '''Get new connection'''
        socket = TSocket.TSocket(self.host, self.port)
        transport = TTransport.TFramedTransport(socket)
        protocol = TBinaryProtocol(transport)
        client = FunicornService.Client(protocol)
        transport.open()
        return client, transport

    def __enter__(self):
        self.client, self.transport = self.get_connection()
        self.logger.debug('Initialize new connection!')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.transport.isOpen():
            self.logger.debug('Connection closed!')
            self.transport.close()

    def preinit_connection(self):
        if self.client is None:
            self.client, self.transport = self.get_connection()

    def predict_img_arr(self, img_arr):
        self.preinit_connection()
        img_bytes = img_arr_to_img_bytes(img_arr)
        return self.client.predict_img_bytes(img_bytes)

    def predict_img_bytes(self, img_bytes):
        self.preinit_connection()
        return self.client.predict_img_bytes(img_bytes)

    def ping(self):
        self.preinit_connection()
        return self.client.ping()

    def close(self):
        if self.transport or self.transport.isOpen():
            self.transport.close()



# import tensorflow as tf
from funicorn import Funicorn
from funicorn.utils import get_args_from_class
import click

def hello():
    print("Hello World")


class HelloClass():
    def call(self):
        print("Hello from HelloClass")


# init_args = get_args_from_class(Funicorn)
# funicorn_options = ['click.option(' + str(('--' + str(arg), 'help=' + str(arg.replace('_', " ")))) for arg in init_args]
# print(funicorn_options)


from setuptools import setup, find_packages
from funicorn import __version__
from os import path

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='funicorn',
    version=__version__,
    description='Funicorn',
    packages=find_packages(),
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    author='congvm',
    author_email='congvm.it@gmail.com',
    license='MIT',
    zip_safe=False,
    install_requires=requirements,
    classifiers=(
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ),
    entry_points={
        'console_scripts': ['funicorn=funicorn.cli:start',
                            'funicorn-terminate=funicorn.cli:worker_terminate',
                            'funicorn-idle=funicorn.cli:worker_idle',
                            'funicorn-resume=funicorn.cli:worker_resume',
                            'funicorn-restart=funicorn.cli:worker_restart',
                            'funicorn-add=funicorn.cli:add_workers',
                            'funicorn-status=funicorn.cli:status',
                            ],
    }
)



import os
import multiprocessing as mp
from random import randint
import threading
from collections import namedtuple
import uuid
import time
import json
import traceback
from queue import Empty
from queue import Queue
from .exceptions import LengthEqualtyError
from .utils import img_bytes_to_img_arr, get_args_from_class
from .logger import get_logger
from .utils import colored_worker_name, colored_funicorn_name, colored_network_name
from .mqueue import Queue as MQueue

MAX_QUEUE_SIZE = 1000
RESULT_TIMEOUT = 0.0001
DEFAULT_TIMEOUT = 500
WORKER_TIMEOUT = 3
DEFAULT_BATCH_TIMEOUT = 0.01


__all__ = ['Funicorn']
Task = namedtuple('Task', ['request_id', 'data'])
WorkerInfo = namedtuple('WorkerInfo', ['wrk', 'wrk_id', 'pid', 'gpu_id',
                                       'ps_status', 'queue',
                                       'ready_event',
                                       'terminate_event'])


class BaseWorker():
    def __init__(self, model_cls, result_dict=None,
                 batch_size=1, batch_timeout=DEFAULT_BATCH_TIMEOUT,
                 ready_event=None, terminate_event=None, model_init_kwargs=None,
                 debug=False):

        self._worker_id = None
        self._model_init_kwargs = model_init_kwargs or {}
        self._model_cls = model_cls
        self._wrk_queue = None
        self._result_dict = result_dict
        self.batch_size = batch_size
        self.batch_timeout = batch_timeout
        self._ready_event = ready_event
        self._terminate_event = terminate_event
        self._pid = os.getpid()
        self._model = None
        self._debug = debug
        self.logger = get_logger(colored_worker_name(
            'BASE-WORKER'), mode='debug' if self._debug else 'info')

    def _recv_request(self):
        raise NotImplementedError

    def _send_response(self, request_id, result):
        raise NotImplementedError

    def _init_environ(self):
        # INFO messages are not printed
        os.environ['TF_CPP_MIN_LOG_LEVEL'] = '1'

    def run_once(self):
        # Get data from queue
        batch = []
        start_time = time.time()
        for _ in range(self.batch_size):
            try:
                task = self._recv_request()
            except TimeoutError:
                break
            else:
                batch.append(task)
        if not batch:
            return 0

        batch_size = len(batch)
        model_input = [task.data for task in batch]
        # Model predict
        start_model_time = time.time()
        results = self._model.predict(model_input)
        end_model_time = time.time()
        assert isinstance(results, list), ValueError(
            '`results` must be a list but receive `{}` which is not valid'.format(results))
        assert len(results) == len(batch), LengthEqualtyError(
            'Length of result and batch must be equal')
        for (task, result) in zip(batch, results):
            self._send_response(task.request_id, result)
        self.logger.debug(
            f'Inference with batch_size: {batch_size} - inference-time: {time.time() - start_time} - model-time: {end_model_time - start_model_time}')
        # Return None or something to notify number of data in queue
        return batch_size

    def run(self):
        '''Loop into a queue'''
        while True:
            try:
                self.logger.info('Process new data!')
                handled = self.run_once()
                if self._ready_event and not self._ready_event.is_set() and (self._wrk_queue.qsize() == 0):
                    self.logger.info('All jobs have been done. Terminated')
                    self._terminate_event.set()
                    break
                if not handled:
                    time.sleep(RESULT_TIMEOUT)
            except Exception as e:
                self.logger.error(e)


class Worker(BaseWorker):
    def _recv_request(self):
        try:
            task = self._wrk_queue.get(timeout=self.batch_timeout)
        except Empty:
            raise TimeoutError
        else:
            return task

    def _send_response(self, request_id, result):
        self._result_dict[request_id] = result

    def run(self, worker_id=None, gpu_id=None, ready_event=None, terminate_event=None, wrk_queue=None):
        ''' Init process parameters
            Every param initialized here are seperable among processes
        '''
        self.logger = get_logger(
            colored_worker_name(f'WORKER-{worker_id}'), mode='debug' if self._debug else 'info')
        self._init_environ()
        self._wrk_queue = wrk_queue
        self._worker_id = worker_id
        self._pid = os.getpid()
        worker_args = get_args_from_class(self._model_cls)
        if 'gpu_id' in worker_args:
            self._model_init_kwargs.update({'gpu_id': gpu_id})
        device = f'GPU-{gpu_id}' if gpu_id else 'CPU'
        self.logger.info(f'Initializing Worker in {device}')
        self._model = self._model_cls(**self._model_init_kwargs)

        if ready_event:
            self._ready_event = ready_event
            self._ready_event.set()  # tell father process that init is finished

        if terminate_event:
            self._terminate_event = terminate_event

        super().run()


class Funicorn():
    '''Lightweight Deep Learning Inference Framework'''

    def __init__(self, model_cls, num_workers=1, batch_size=1, batch_timeout=10,
                 max_queue_size=1000,
                 gpu_devices=None,
                 model_init_kwargs=None, debug=False, timeout=5000):
        self.model_cls = model_cls
        self.logger = get_logger(
            colored_funicorn_name(), mode='debug' if debug else 'info')
        self._model_init_kwargs = model_init_kwargs or {}
        self.timeout = timeout
        self.debug = debug
        self._lock = threading.Lock()
        self.gpu_devices = gpu_devices
        self.num_workers = num_workers

        self._input_queue = MQueue()
        self._result_dict = mp.Manager().dict()
        if batch_size == 1:
            batch_timeout = None
        elif batch_timeout is not None:
            batch_timeout = batch_timeout/1000
        self._wrk = Worker(self.model_cls, self._result_dict,
                           batch_size=batch_size, batch_timeout=batch_timeout,
                           model_init_kwargs=model_init_kwargs, debug=self.debug)

        self.pid = os.getpid()
        # self._init_stat()
        self.idle_event = mp.Event()
        self.wrk_ps = []
        self.connection_apps = {}

    def register_connection(self, connection):
        self.logger.info(
            f'Register {colored_network_name(connection.name)} connection')
        self.connection_apps[connection.name] = connection

    def get_worker_pids(self):
        return [wrk.pid for wrk in self.wrk_ps]

    def _init_connections(self):
        for conn_name, conn in self.connection_apps.items():
            conn.start()

    def _init_all_workers(self):
        if self.model_cls is None:
            self.logger.warning(
                'Cannot start workers because model class is not provided!')
        else:
            if self.num_workers <= 0:
                self.logger.warning(
                    'Cannot start workers because num workers is set to 0!')
            else:
                self.add_worker(self.num_workers, self.gpu_devices)

    def _wait_for_worker(self, timeout=WORKER_TIMEOUT):
        # wait for all workers finishing init
        for (i, worker_info) in enumerate(self.wrk_ps):
            # todo: select all events with timeout
            is_ready = worker_info.ready_event.wait(timeout)
            self.logger.info(
                f"{colored_worker_name(f'WORKER-{worker_info.wrk_id}')} ready state: {is_ready}")
            if not is_ready:
                self.logger.error(
                    f"{colored_worker_name(f'WORKER-{worker_info.wrk_id}')} cannot start. EXIT!")
                exit()

    def _start_task_distributations(self):
        '''Distribute task to wrk_queue'''
        while True:
            task = self._input_queue.get()
            self.logger.debug(
                f'Get data from input queue: {self._input_queue}')
            with self._lock:
                input_queue = self.wrk_ps[randint(
                    0, len(self.wrk_ps) - 1)].queue
            input_queue.put(task)

    def predict(self, data, asynchronous=False):
        '''Main function to predict data'''
        request_id = str(uuid.uuid4())
        self._input_queue.put(Task(request_id=request_id, data=data))
        self.logger.info(
            f'Received data with request_id: {request_id}')
        if asynchronous:
            return request_id
        else:
            return self.get_result(request_id)

    @property
    def input_queue(self):
        self.logger.info(f'Get input queue: {self._input_queue}')
        return self._input_queue

    @property
    def result_dict(self):
        return self._result_dict

    def get_result(self, request_id):
        ret = None
        while True:
            ret = self._result_dict.pop(request_id, None)
            if ret is not None:
                break
            time.sleep(RESULT_TIMEOUT)
        self.logger.debug(f'Sent result of request_id to client: {request_id}')
        return ret

    def add_more_workers(self, num_workers, gpu_devices):
        num_workers = int(num_workers)
        device = 'CPU' if gpu_devices is None else gpu_devices
        self.logger.info(f'Add {num_workers} workers in {device}')
        self.num_workers += num_workers
        self.add_worker(num_workers, gpu_devices)
        return 'Added more workers!'

    def add_worker(self, num_workers, gpu_devices):
        for idx in range(num_workers):
            ready_event = mp.Event()
            terminate_event = mp.Event()
            args = (ready_event, terminate_event)
            if gpu_devices is not None:
                gpu_id = gpu_devices[idx % len(gpu_devices)]
            else:
                gpu_id = None
            wrk_queue = MQueue()  # mp.Queue()
            worker_id = randint(0, 999999)
            args = (worker_id, gpu_id, ready_event,
                    terminate_event, wrk_queue)
            wrk = mp.Process(target=self._wrk.run, args=args,
                             daemon=True,
                             name=f'funicorn-worker-{worker_id}')
            wrk.start()
            worker_info = WorkerInfo(wrk=wrk,
                                     wrk_id=worker_id,
                                     pid=wrk.pid,
                                     gpu_id=gpu_id,
                                     ps_status='unknown',
                                     queue=wrk_queue,
                                     ready_event=ready_event,
                                     terminate_event=terminate_event)
            with self._lock:
                self.wrk_ps.append(worker_info)

    def terminate_all_workers(self):
        '''Terminate all workers'''
        if len(self.wrk_ps) == 0:
            self.logger.info('No workers running')
        else:
            self.logger.info(
                'Received TERMINATE signal. All workers will be killed soon when they finish their jobs')

        for worker_info in self.wrk_ps:
            worker_info.ready_event.clear()

        terminate_workers = []
        for (i, worker_info) in enumerate(self.wrk_ps):
            is_terminated = worker_info.terminate_event.wait(20000)
            if is_terminated:
                self.wrk_ps.remove(worker_info)
                terminate_workers.append(worker_info)
        return f'Processes will be killed: {", ".join([str(worker_info.pid) for worker_info in terminate_workers])} and there is/are {len(self.wrk_ps)} left'

    def idle_all_workers(self):
        '''Idle all workers'''
        self.logger.info(
            'Received IDLE signal. All workers will idle until receiving resume signal')
        self.idle_event.set()
        return 'All workers are idling!'

    def resume_all_workers(self):
        self.logger.info(
            'Received RESUME signal. All workers will idle until receiving resume signal')
        self.idle_event.clear()
        return 'All workers are resuming!'

    def restart_all_workers(self):
        self.idle_all_workers()
        self.terminate_all_workers()
        self.logger.info('RESTART all workers')
        self._init_all_workers()
        self._wait_for_worker()
        self.resume_all_workers()
        return 'All workers are restarted!'

    def check_all_worker(self):
        '''Check status of all workers. Restart them if necessary'''
        pass

    def _recheck_all_modules(self):
        if len(self.connection_apps) == 0 and not self.model_cls:
            self.logger.error("Nothing is running. STOP!")
            exit()

    def _serve(self):
        try:
            self._init_all_workers()
            self._wait_for_worker()
            self._init_connections()
            self._recheck_all_modules()
            self._start_task_distributations()
        except KeyboardInterrupt:
            exit()
        except Exception as e:
            self.logger.error(traceback.format_exc())

    def serve(self, run_in_background=False):
        if run_in_background:
            t = threading.Thread(target=self._serve, daemon=True)
            t.start()
        else:
            self._serve()
